// HuffYUV
//
// Changes copyright 2003 Klaus Post, and distributed under
// the terms of the GNU General Public License, v2 or later.  See
// http://www.gnu.org/copyleft/gpl.html.
//
//
//Pixels referenced:
// 12
// 34

void isse_reduceby2_32(const unsigned char *_in, unsigned char *_out, int out_w, int out_h, int instride, int outstride) {
  int instride2= instride*2;
  __asm {
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [out_h];  // Height
    mov edx,[instride];
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0,[esi+ecx*8]  // Load pix 1
    movd mm1,[esi+ecx*8+4]  // Load pix 2
    add esi,edx;
    movd mm2,[esi+ecx*8]  // Load pix 3
    movd mm3,[esi+ecx*8+4]  // Load pix 4
    pavgb mm0,mm1
    pavgb mm2,mm3
    sub esi,edx
    pavgb mm0,mm2
    movd [edi+ecx*4],mm0
    inc ecx

    cmp  ecx,[out_w]
    jl loopback_x
    
    add edi,[outstride]
    add esi,[instride2]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}


void isse_reduceby2_24(const unsigned char *_in, unsigned char *_out, int out_w, int out_h, int instride, int outstride) {
  int instride2= instride*2;
  out_w *=3;
  __asm {
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [out_h];  // Height
    mov edx,[instride];
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0,[esi+ecx*2]  // Load pix 1
    movd mm1,[esi+ecx*2+3]  // Load pix 2
    add esi,edx;
    movd mm2,[esi+ecx*2]  // Load pix 3
    movd mm3,[esi+ecx*2+3]  // Load pix 4
    pavgb mm0,mm1
    pavgb mm2,mm3
    sub esi,edx
    pavgb mm0,mm2
    movd [edi+ecx],mm0
    add ecx,3

    cmp  ecx,[out_w]
    jl loopback_x
    
    add edi,[outstride]
    add esi,[instride2]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}



void isse_upscaleby2_32(const unsigned char *_in, unsigned char *_out, int in_w, int in_h, int instride, int outstride) {
  int outstride2 = outstride*2;
  __asm {
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [in_h];  // Height
    mov edx, [outstride];
    mov ebx, [in_w]
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0,[esi+ecx*4]  // Load pix 
    pshufw mm0,mm0, 01000100b
    movq [edi+ecx*8],mm0  // Store pix 1 & 2
    add edi,edx;
    movq [edi+ecx*8],mm0  // Store pix 1 & 2
    sub edi,edx
    inc ecx

    cmp  ecx,ebx
    jl loopback_x
    
    add edi,[outstride2]
    add esi,[instride]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}

void isse_reduceby2_yuy2(const unsigned char *_in, unsigned char *_out, int out_w, int out_h, int instride, int outstride) {
  _declspec(align(8)) const __int64 mask = 0x00000000ff00ffff;
  _declspec(align(8)) const __int64 inv_mask = 0x0000000000ff0000;
  int instride2= instride*2;

  __asm {
    movq mm7, [mask]
    movq mm6, [inv_mask]
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [out_h];  // Height
    mov edx,[instride];
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0,[esi+ecx*4]  // Load pix 1 (We have luma1 + chroma from this)
    movd mm1,[esi+ecx*4+4]  // Load next pixel  (we take luma 2 from this)
    add esi,edx;
    movd mm2,[esi+ecx*4]  // Load pix 1 (We have luma1 + chroma from this)
    movd mm3,[esi+ecx*4+4]  // Load next pixel  (we take luma 2 from this)
    pavgb mm0,mm2
    pavgb mm1,mm3
    
    pand mm0,mm7
    pand mm1,mm6
    por mm0,mm1
    sub esi,edx
    movd [edi+ecx*2],mm0
    add ecx,2

    cmp  ecx,[out_w]
    jl loopback_x
    
    add edi,[outstride]
    add esi,[instride2]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}


void mmx_reduceby2_yuy2(const unsigned char *_in, unsigned char *_out, int out_w, int out_h, int instride, int outstride) {
  _declspec(align(8)) const __int64 mask = 0x00000000ff00ffff;
  _declspec(align(8)) const __int64 inv_mask = 0x0000000000ff0000;
  int instride2= instride*2;
  __asm {
    movq mm7, [mask]
    movq mm6, [inv_mask]
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [out_h];  // Height
    mov edx,[instride];
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0,[esi+ecx*4]  // Load pix 1 (We have luma1 + chroma from this)
    movd mm1,[esi+ecx*4+4]  // Load next pixel  (we take luma 2 from this)
    
    pand mm0,mm7
    pand mm1,mm6
    por mm0,mm1

    movd [edi+ecx*2],mm0
    add ecx,2

    cmp  ecx,[out_w]
    jl loopback_x
    
    add edi,[outstride]
    add esi,[instride2]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}


void mmx_upscaleby2_24(const unsigned char *_in, unsigned char *_out, int in_w, int in_h, int instride, int outstride) {
  int outstride2 = outstride*2;
  in_w*=3;
  __asm {
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [in_h];  // Height
    mov edx,[outstride];
    mov ebx, [in_w]
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0,[esi+ecx]  // Load pix 
    movd [edi+ecx*2],mm0  // Store pix 1 
    movd [edi+ecx*2+3],mm0  // Store pix 2
    add edi,edx;
    movd [edi+ecx*2],mm0  // Store pix 3
    movd [edi+ecx*2+3],mm0  // Store pix 4
    sub edi,edx
    add ecx,3

    cmp  ecx, ebx
    jl loopback_x
    
    add edi,[outstride2]
    add esi,[instride]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}

void mmx_upscaleby2_32(const unsigned char *_in, unsigned char *_out, int in_w, int in_h, int instride, int outstride) {
  in_w*=4;
  int outstride2 = outstride*2;
  __asm {
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [in_h];  // Height
    mov edx,[outstride];
    mov ebx, [in_w]
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0,[esi+ecx]  // Load pix 
    movd [edi+ecx*2],mm0  // Store pix 1 
    movd [edi+ecx*2+4],mm0  // Store pix 2
    add edi,edx;
    movd [edi+ecx*2],mm0  // Store pix 3
    movd [edi+ecx*2+4],mm0  // Store pix 4
    sub edi,edx
    add ecx,4

    cmp  ecx,ebx
    jl loopback_x
    
    add edi,[outstride2]
    add esi,[instride]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}

void isse_upscaleby2_yuy2(const unsigned char *_in, unsigned char *_out, int in_w, int in_h, int instride, int outstride) {
  in_w*=2;
  int outstride2 = outstride*2;
  __asm {
    pxor mm3,mm3
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [in_h];  // Height
    mov edx,[outstride];
    mov ebx, [in_w]
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0, [esi+ecx]
    punpcklbw mm0, mm3  // Load pix (Expands to 4x2 pixels)   
    pshufw mm1,mm0, 11000100b  // First pix-set (also written below)
    pshufw mm2,mm0, 11100110b  // Second pix-set
    
    packuswb mm1,mm3
    packuswb mm2,mm3

    movd [edi+ecx*2],mm1  // Store pix-set 1 
    movd [edi+ecx*2+4],mm2 // Store pix-set 2
    add edi,edx;
    movd [edi+ecx*2],mm1    // Store pix-set 3
    movd [edi+ecx*2+4],mm2  // Store pix-set 4

    sub edi,edx
    add ecx,4

    cmp  ecx,ebx
    jl loopback_x
    
    add edi,[outstride2]
    add esi,[instride]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}

void mmx_upscaleby2_yuy2(const unsigned char *_in, unsigned char *_out, int in_w, int in_h, int instride, int outstride) {
  in_w*=2;
  int outstride2 = outstride*2;
  __asm {
    pxor mm3,mm3
    mov esi, [_in];
    mov edi, [_out];
    mov eax, [in_h];  // Height
    mov edx,[outstride];
    mov ebx, [in_w]
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movd mm0, [esi+ecx]
    punpcklbw mm0, mm3  // Load pix (Expands to 4x2 pixels)   
    movq mm1,mm0
    movq mm2,mm0
    
    packuswb mm1,mm3
    packuswb mm2,mm3

    movd [edi+ecx*2],mm1  // Store pix-set 1 
    movd [edi+ecx*2+4],mm2 // Store pix-set 2
    add edi,edx;
    movd [edi+ecx*2],mm1    // Store pix-set 3
    movd [edi+ecx*2+4],mm2  // Store pix-set 4

    sub edi,edx
    add ecx,4

    cmp  ecx,ebx
    jl loopback_x
    
    add edi,[outstride2]
    add esi,[instride]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}

void mmx_removeAlpha(unsigned char *_in, int in_w, int in_h, int instride) {
  in_w*=4;  // BPP
  _declspec(align(8)) const __int64 mask = 0x00ffffff00ffffff;
  __asm {
    movq mm7,[mask]
    mov esi, [_in]
    mov eax, [in_h]  // Height
    mov edx,[instride]
    mov edi, [in_w]
    align 16
loopback_y:
    xor ecx,ecx; // Width
    align 16
loopback_x:
    movq mm0,[esi+ecx]  // Load pix 
    pand mm0,mm7
    movq [esi+ecx],mm0
    add ecx,8

    cmp  ecx, edi
    jl loopback_x
    
    add esi,[instride]
    dec eax
    jnz loopback_y
  }
  __asm {emms} 
}
